@@ -1,10 +1,19 @@
+import functools
 import logging
 from io import TextIOBase
-from typing import Callable, TypeVar
+from typing import Callable, List, Optional, Tuple, TypeVar, Union
 
-from simplini.core import IniConfigBase, IniConfigOption, IniConfigSection
+from simplini.core import (
+    IniConfigBase,
+    IniConfigOption,
+    IniConfigSection,
+    IniFlavour,
+    SimpliniError,
+    ValuePresentationStyle,
+)
 
 T = TypeVar("T")
+ParseFn = Callable[[], T]
 
 
 LOGGER = logging.getLogger(__name__)
@@ -16,18 +25,23 @@ class PositionContext:
         line: str,
         line_number: int,
         column_number: int,
+        lines_before: Optional[List[str]] = None,
     ):
         self.line = line
+        # 1-based line number of the line
         self.line_number = line_number
+        # 1-based position in the current line
         self.column_number = column_number
+        # zero or multiple lines before the current line
+        self.lines_before = lines_before
 
 
-class ParsingError(Exception):
+class ParsingError(SimpliniError):
     def __init__(self, message: str) -> None:
         super().__init__(message)
         self.message: str = message
-        self.position: int | None = None
-        self.position_context: PositionContext | None = None
+        self.position: Optional[int] = None
+        self.position_context: Optional[PositionContext] = None
 
     def extend_message(self, message: str) -> None:
         self.message += message
@@ -37,7 +51,7 @@ class ParsingError(Exception):
 class RecursiveDescentParserBase:
     def __init__(self, text_io: TextIOBase):
         self.text_io: TextIOBase = text_io
-        self.deepest_error: ParsingError | None = None
+        self.deepest_error: Optional[ParsingError] = None
 
     def parsing_error(self, message: str) -> ParsingError:
         position = self.text_io.tell()
@@ -45,28 +59,47 @@ class RecursiveDescentParserBase:
         error = ParsingError(message)
         error.position = position
 
-        # update deepest error
-        if self.deepest_error is None or position > self.deepest_error.position:
+        # update the deepest error
+        # if the position is the same as the deepest update the message too
+        # as it might provide an enhanced wording
+        if self.deepest_error is None or position >= self.deepest_error.position:
             self.deepest_error = error
 
         return error
 
-    def expect(self, expected: str) -> str:
-        char = self.text_io.read(1)
+    @staticmethod
+    def represent(char: str) -> str:
+        if char == "\n":
+            return "LF"
+        elif char == "\r":
+            return "CR"
+        elif char == "":
+            return "EOF"
+        else:
+            return f'"{char}"'
 
-        if char != expected:
-            if char == "\n":
-                actual = "LF"
-            elif char == "\r":
-                actual = "CR"
-            elif char == "":
-                actual = "EOF"
-            else:
-                actual = f'"{char}"'
+    def expect(
+        self,
+        expected: Union[str, List[str]],
+        error: Optional[str] = None,
+    ) -> str:
+        actual = self.text_io.read(1)
 
-            raise self.parsing_error(f'Expected "{expected}", but encountered {actual}')
+        if isinstance(expected, str):
+            expected = [expected]
 
-        return char
+        for value in expected:
+            if actual == value:
+                break
+        else:
+            if not error:
+                error = (
+                    f"Expected {' OR '.join(self.represent(v) for v in expected)}, "
+                    f"but encountered {self.represent(actual)}"
+                )
+            raise self.parsing_error(error)
+
+        return actual
 
     def expect_eof(self) -> None:
         char = self.text_io.read(1)
@@ -77,8 +110,9 @@ class RecursiveDescentParserBase:
         return None
 
     def accept(
-        self, value_or_predicate: Callable[[str], bool] | str
-    ) -> tuple[bool, str | None]:
+        self,
+        value_or_predicate: Union[Callable[[str], bool], str],
+    ) -> Tuple[bool, Optional[str]]:
         char = self.text_io.read(1)
 
         # reached EOF
@@ -101,8 +135,8 @@ class RecursiveDescentParserBase:
 
     def accept_multiple(
         self,
-        value_or_predicate: Callable[[str], bool] | str,
-    ) -> tuple[bool, str]:
+        value_or_predicate: Union[Callable[[str], bool], str],
+    ) -> Tuple[bool, str]:
         chars = ""
         accepted, char = self.accept(value_or_predicate)
 
@@ -116,7 +150,7 @@ class RecursiveDescentParserBase:
 
         return True, chars
 
-    def multiple(self, parse_fn: Callable[[], T]) -> list[T]:
+    def multiple(self, parse_fn: ParseFn) -> List[T]:
         results = []
 
         while True:
@@ -130,7 +164,7 @@ class RecursiveDescentParserBase:
 
         return results
 
-    def optional(self, parse_fn: Callable[[], T]) -> tuple[bool, T | None]:
+    def optional(self, parse_fn: ParseFn) -> Tuple[bool, Optional[T]]:
         position = self.text_io.tell()
 
         try:
@@ -140,7 +174,7 @@ class RecursiveDescentParserBase:
             self.text_io.seek(position)
             return False, None
 
-    def choice(self, parse_fns: list[Callable[[], T]]) -> tuple[int, T]:
+    def choice(self, parse_fns: List[ParseFn]) -> Tuple[int, T]:
         last_error = None
         position = self.text_io.tell()
 
@@ -162,21 +196,26 @@ class RecursiveDescentParserBase:
         return peeked
 
     def hinted_choice(
-        self, hinted_parse_fns: list[tuple[str | None, Callable[[], T]]]
-    ) -> T:
+        self,
+        hinted_parse_fns: List[Tuple[Union[Optional[str], List[str]], ParseFn]],
+    ) -> Tuple[int, T]:
         position = self.text_io.tell()
         last_error = None
 
-        for hint, parse_fn in hinted_parse_fns:
+        for parser_idx, (hint, parse_fn) in enumerate(hinted_parse_fns):
             if hint:
-                peeked = self.peek(len(hint))
+                if isinstance(hint, str):
+                    hint = [hint]
 
-                # hint matched, so we know this branch is what we need
-                if hint == peeked:
-                    return parse_fn()
+                for single_hint_value in hint:
+                    peeked = self.peek(len(single_hint_value))
+
+                    # hint matched, so we know this branch is what we need
+                    if single_hint_value == peeked:
+                        return parser_idx, parse_fn()
             else:  # empty hint, attempt the branch and go to next if not successful
                 try:
-                    return parse_fn()
+                    return parser_idx, parse_fn()
                 except ParsingError as e:
                     self.text_io.seek(position)
                     last_error = e
@@ -185,57 +224,42 @@ class RecursiveDescentParserBase:
         raise last_error
 
 
-class RecursiveDescentIniParserImpl(RecursiveDescentParserBase):
+class IniParserImpl(RecursiveDescentParserBase):
     def __init__(
         self,
         text_io: TextIOBase,
-        allow_unquoted_values: bool,
-        key_value_separator: str,
-        comment_separator: str,
-        escape_character: str,
-        quote_character: str,
-        escape_sequences: dict[str, str],
-        new_line: str,
+        flavour: IniFlavour,
     ):
         super().__init__(text_io)
-        self.allow_unquoted_values = allow_unquoted_values
-        self.key_value_separator = key_value_separator
-        self.comment_separator = comment_separator
-        self.escape_character = escape_character
-        self.quote_character = quote_character
-        self.escape_sequences = escape_sequences
-        self.new_line = new_line
+        self.flavour = flavour
 
     def resolve_escape_sequence(self, sequence: str) -> str:
-        if sequence not in self.escape_sequences:
+        if sequence not in self.flavour.escape_sequences:
             raise self.parsing_error(f"Unknown escape sequence: {sequence}")
-        return self.escape_sequences[sequence]
+        return self.flavour.escape_sequences[sequence]
 
     def parse_quoted_string(self) -> str:
         self.parse_whitespaces()
-
-        self.expect(self.quote_character)
+        self.expect(self.flavour.quote_character)
 
         value = ""
 
         while True:
             char = self.text_io.read(1)
-
-            if char == self.escape_character:
+            if char == self.flavour.escape_character:
                 next_char = self.text_io.read(1)
-                value += self.resolve_escape_sequence(char + next_char)
-            elif char == self.new_line:
+                value += self.resolve_escape_sequence(next_char)
+            elif char == self.flavour.new_line:
                 raise self.parsing_error(
                     "New line encountered before closing quoted string"
                 )
             elif char == "":
                 raise self.parsing_error("EOF encountered before closing quoted string")
             else:  # normal character
-                if char == self.quote_character:
+                if char == self.flavour.quote_character:
                     break
                 value += char
 
-        # trailing whitespaces
         self.parse_whitespaces()
 
         return value
@@ -243,25 +267,30 @@ class RecursiveDescentIniParserImpl(RecursiveDescentParserBase):
     def parse_triple_quoted_string(self) -> str:
         self.parse_whitespaces()
 
-        self.expect(self.quote_character)
-        self.expect(self.quote_character)
-        self.expect(self.quote_character)
+        self.expect(self.flavour.quote_character)
+        self.expect(self.flavour.quote_character)
+        self.expect(self.flavour.quote_character)
 
         value = ""
 
         while True:
             char = self.text_io.read(1)
 
-            if char == self.escape_character:
+            if char == self.flavour.escape_character:
                 next_char = self.text_io.read(1)
-                value += self.resolve_escape_sequence(char + next_char)
+                value += self.resolve_escape_sequence(next_char)
+            elif char == "":
+                # encountered EOF
+                raise self.parsing_error(
+                    "EOF encountered before closing triple quoted string"
+                )
             else:  # normal character:
-                if char == self.quote_character:
+                if char == self.flavour.quote_character:
                     # check if it's the end of the triple-quoted string
                     next_char = self.text_io.read(1)
-                    if next_char == self.quote_character:
+                    if next_char == self.flavour.quote_character:
                         next_next_char = self.text_io.read(1)
-                        if next_next_char == self.quote_character:
+                        if next_next_char == self.flavour.quote_character:
                             break
                         else:
                             value += char + next_char + next_next_char
@@ -270,7 +299,6 @@ class RecursiveDescentIniParserImpl(RecursiveDescentParserBase):
                 else:
                     value += char
 
-        # trailing whitespaces
         self.parse_whitespaces()
 
         return value
@@ -278,32 +306,38 @@ class RecursiveDescentIniParserImpl(RecursiveDescentParserBase):
     def parse_unquoted_string(self) -> str:
         self.parse_whitespaces()
 
-        _, value = self.accept_multiple(
-            lambda c: c not in (self.new_line, self.comment_separator),
-            # even if nothing is accepted we consider the value to be empty
-        )
+        def is_acceptable(c: str) -> bool:
+            if not self.flavour.allow_inline_comments:
+                return c not in self.flavour.new_line
+            return c not in (
+                (self.flavour.new_line,) + tuple(self.flavour.comment_markers)
+            )
+
+        _, value = self.accept_multiple(is_acceptable)
 
         assert value is not None
 
         # strip trailing spaces
         value = value.rstrip()
 
-        if self.quote_character in value:
+        if self.flavour.quote_character in value:
             raise self.parsing_error(
                 "Quote character inside non-quoted strings is forbidden as ambiguous"
             )
 
-        # trailing whitespaces
         self.parse_whitespaces()
 
         return value
 
-    def parse_option_name(self) -> str:
-        def is_option_name_char(c: str) -> bool:
-            return c.isalnum() or c in ("_", "-", ".", ":")
+    def is_option_name_char(self, c: str) -> bool:
+        # here is what TOML allows: A-Za-z0-9_-
+        # see https://toml.io/en/v1.0.0
+        # we will allow a bit more for now
+        return c.isalnum() or c in ("_", "-", ".", ":")
 
+    def parse_option_name(self) -> str:
         _, option_name = self.accept_multiple(
-            is_option_name_char,
+            self.is_option_name_char,
         )
 
         if not option_name:
@@ -311,35 +345,51 @@ class RecursiveDescentIniParserImpl(RecursiveDescentParserBase):
 
         return option_name
 
-    def parse_option_value(self) -> str:
-        if self.allow_unquoted_values:
-            return self.hinted_choice(
+    def parse_option_value(self) -> Tuple[str, ValuePresentationStyle]:
+        if self.flavour.allow_unquoted_values:
+            parser_idx, value = self.hinted_choice(
                 [
-                    (self.quote_character * 3, self.parse_triple_quoted_string),
-                    (self.quote_character, self.parse_quoted_string),
+                    (self.flavour.quote_character * 3, self.parse_triple_quoted_string),
+                    (self.flavour.quote_character, self.parse_quoted_string),
                     (None, self.parse_unquoted_string),
                 ]
             )
+
+            if parser_idx == 0:
+                return value, ValuePresentationStyle.TRIPLE_QUOTED
+            elif parser_idx == 1:
+                return value, ValuePresentationStyle.QUOTED
+            elif parser_idx == 2:
+                return value, ValuePresentationStyle.UNQUOTED
+            else:
+                raise NotImplementedError(parser_idx)
         else:
-            return self.hinted_choice(
+            parser_idx, value = self.hinted_choice(
                 [
-                    (self.quote_character * 3, self.parse_triple_quoted_string),
+                    (self.flavour.quote_character * 3, self.parse_triple_quoted_string),
                     (
-                        self.quote_character,
+                        self.flavour.quote_character,
                         self.parse_quoted_string,
                     ),
                 ]
             )
 
+            if parser_idx == 0:
+                return value, ValuePresentationStyle.TRIPLE_QUOTED
+            elif parser_idx == 1:
+                return value, ValuePresentationStyle.QUOTED
+            else:
+                raise NotImplementedError(parser_idx)
+
     def parse_option(self) -> IniConfigOption:
         comments = self.parse_comments()
         self.parse_whitespaces()
         option_name = self.parse_option_name()
         self.parse_whitespaces()
-        self.expect(self.key_value_separator)
+        self.expect(self.flavour.key_value_separators)
         self.parse_whitespaces()
 
-        option_value = self.parse_option_value()
+        option_value, option_value_style = self.parse_option_value()
 
         _, inline_comment = self.optional(self.parse_comment_line)
         self.multiple(self.parse_empty_line)
@@ -347,10 +397,11 @@ class RecursiveDescentIniParserImpl(RecursiveDescentParserBase):
         option = IniConfigOption(option_name, option_value)
         option.comment = comments
         option.inline_comment = inline_comment
+        option.style = option_value_style
         return option
 
     def is_whitespace(self, char: str) -> bool:
-        return char in (" ", "\t")
+        return char in self.flavour.whitespace_characters
 
     def parse_whitespaces(self) -> str:
         return self.accept_multiple(self.is_whitespace)[1]
@@ -370,25 +421,25 @@ class RecursiveDescentIniParserImpl(RecursiveDescentParserBase):
 
     def parse_comment_line(self) -> str:
         self.accept_multiple(self.is_whitespace)
-        self.expect(self.comment_separator)
+        self.expect(self.flavour.comment_markers)
 
         _, comment = self.accept_multiple(
-            lambda c: c != "\n",
+            lambda c: c != self.flavour.new_line,
         )
 
         # accept new line if present as well
-        self.accept("\n")
+        self.accept(self.flavour.new_line)
 
         # strip leading/trailing spaces
         return comment.strip()
 
     def parse_empty_line(self) -> str:
         self.accept_multiple(self.is_whitespace)
-        self.expect(self.new_line)
+        self.expect(self.flavour.new_line)
         return ""
 
-    def parse_comments(self) -> list[str]:
-        def parse_comment_or_empty_line() -> tuple[int, str]:
+    def parse_comments(self) -> List[str]:
+        def parse_comment_or_empty_line() -> Tuple[int, str]:
             parser_idx, parsed_value = self.choice(
                 [
                     self.parse_comment_line,
@@ -412,46 +463,62 @@ class RecursiveDescentIniParserImpl(RecursiveDescentParserBase):
         self.expect("[")
 
         _, section_name = self.accept_multiple(
-            lambda c: c not in ("]", "\n"),
+            lambda c: c not in ("]", self.flavour.new_line),
         )
 
         if not section_name:
             raise self.parsing_error("Expected section name to be non-empty")
 
         self.expect("]")
+        self.accept_multiple(self.is_whitespace)
+
+        alt_idx, result = self.hinted_choice(
+            [
+                (self.flavour.comment_markers, self.parse_comment_line),
+                (
+                    None,
+                    lambda: self.expect(
+                        self.flavour.new_line,
+                        "Expected end of line after section header",
+                    ),
+                ),
+            ]
+        )
 
         section = IniConfigSection(section_name)
+
+        if alt_idx == 0:
+            section.inline_comment = result
+
         section.comment = comments
 
         self.parse_section_body(section)
 
         return section
 
-    def parse(self, config: IniConfigBase):
-        # first comments need a little bit edge-case processing because of the
-        # inherent ambiguity as they apply to the:
-        # * next option if it exists,
-        # * OR to the next section if it exists,
-        # * OR to the comment for unnamed section otherwise;
-        first_comments = self.parse_comments()
+    def parse_comments_only_document_edge_case(self, config: IniConfigBase):
+        comment = self.parse_comments()
 
-        # parse default section
+        self.accept_multiple(self.is_whitespace)
+        self.expect_eof()
+
+        config.trailing_comment = comment
+
+        return config
+
+    def parse_normal(self, config: IniConfigBase):
+        # parse unnamed section
         config.unnamed_section = IniConfigSection(None)
+
         self.parse_section_body(config.unnamed_section)
 
+        if not self.flavour.allow_unnamed_section:
+            if config.unnamed_section.options:
+                raise self.parsing_error("Unnamed section is not allowed")
+
         # then any number of other sections
         sections = self.multiple(self.parse_section)
 
-        # attribute first comments
-        if config.unnamed_section.options:
-            first_option = next(iter(config.unnamed_section.options.values()))
-            first_option.comment = first_comments
-        elif sections:
-            first_section = sections[0]
-            first_section.comment = first_comments
-        else:
-            config.unnamed_section.comment = first_comments
-
         for section in sections:
             assert section is not None
 
@@ -464,67 +531,72 @@ class RecursiveDescentIniParserImpl(RecursiveDescentParserBase):
 
         config.trailing_comment = self.parse_comments()
 
+        self.accept_multiple(self.is_whitespace)
         self.expect_eof()
 
         return config
 
+    def parse(self, config: IniConfigBase):
+        self.choice(
+            [
+                functools.partial(self.parse_comments_only_document_edge_case, config),
+                functools.partial(self.parse_normal, config),
+            ]
+        )
+
 
 class IniParser:
-    def __init__(self):
-        self.allow_unquoted_values = True
-        self.key_value_separator = "="
-        self.comment_separator = "#"
-        self.escape_character = "\\"
-        self.quote_character = '"'
-        self.escape_sequences = {
-            r"\n": "\n",
-            r"\t": "\t",
-            r"\\": "\\",
-            r"\"": '"',
-        }
-        self.new_line = "\n"
-
     @staticmethod
-    def position_context(text_io: TextIOBase, position: int) -> PositionContext | None:
+    def position_context(
+        text_io: TextIOBase,
+        position: int,
+        context_lines: int = 1,
+    ) -> Optional[PositionContext]:
         fd = text_io.fileno()
 
         # re-open file w/o automatic new-line conversion so that
         # position is relevant on platforms where new lines are of different
         # len from Unix style (LF)
         with open(
-            fd, mode="r", encoding=text_io.encoding, newline="", closefd=False
+            fd,
+            mode="r",
+            encoding=text_io.encoding,
+            newline="",
+            closefd=False,
         ) as raw:
             raw.seek(0)
 
-            read = 0
-            line_number = 0
-            prev_line = ""
+            read_bytes = 0
+            lines = raw.readlines()
 
-            while True:
-                line_number += 1
-                line = raw.readline()
+            for line_idx, line in enumerate(lines):
+                line_size_bytes = len(line.encode(text_io.encoding))
+                if position < read_bytes + line_size_bytes:
+                    column_idx = 0
+                    col_bytes = 0
+
+                    # we found the line, now find which character specifically
+                    while position > read_bytes + col_bytes + 1:
+                        col_bytes += len(line[column_idx].encode(text_io.encoding))
+                        column_idx += 1
 
-                if position < read + len(line):
                     return PositionContext(
                         line=line,
-                        line_number=line_number,
-                        column_number=position - read + 1,
+                        line_number=line_idx + 1,
+                        column_number=column_idx + 1,
+                        lines_before=lines[line_idx - context_lines : line_idx],
                     )
 
-                read += len(line)
-
-                if not line:
-                    break
-
-                prev_line = line
+                read_bytes += line_size_bytes
 
             # edge case -- position points right after
             # the last character of the text
-            if position == read:
+            if position == read_bytes:
                 return PositionContext(
-                    line=prev_line,
-                    line_number=line_number - 1,
-                    column_number=len(prev_line),
+                    line=lines[-1],
+                    line_number=line_idx + 1,
+                    column_number=len(lines[-1]),
+                    lines_before=lines[-1 - context_lines : -1],
                 )
 
             # unable to determine
@@ -534,16 +606,11 @@ class IniParser:
         self,
         text_io: TextIOBase,
         instance: IniConfigBase,
+        flavour: IniFlavour,
     ) -> None:
-        parser = RecursiveDescentIniParserImpl(
+        parser = IniParserImpl(
             text_io,
-            allow_unquoted_values=self.allow_unquoted_values,
-            key_value_separator=self.key_value_separator,
-            comment_separator=self.comment_separator,
-            escape_character=self.escape_character,
-            quote_character=self.quote_character,
-            escape_sequences=self.escape_sequences,
-            new_line=self.new_line,
+            flavour=flavour,
         )
 
         try:
@@ -561,20 +628,25 @@ class IniParser:
             if position_context:
                 effective.position_context = position_context
 
-                quotation_prefix = "> "
+                error_marker = "> "
+                padding = " " * len(error_marker)
 
-                position_message = (
-                    f"Line {position_context.line_number}, "
-                    f"Column {position_context.column_number}, "
-                    f"Position {effective.position}\n"
-                )
-                position_message += quotation_prefix
+                position_message = padding + "..." + "\n"
+                for line_before in position_context.lines_before:
+                    position_message += padding + line_before
+
+                position_message += error_marker
                 position_message += position_context.line.rstrip("\n") + "\n"
                 position_message += " " * (
-                    position_context.column_number - 1 + len(quotation_prefix)
+                    position_context.column_number - 1 + len(error_marker)
+                )
+                position_message += "^\n"
+
+                position_message += (
+                    f"Line {position_context.line_number}, "
+                    f"Column {position_context.column_number}, "
+                    f"Byte {effective.position}\n"
                 )
-                position_message += "^"
-                position_message += "\n"
 
                 effective.extend_message("\n\n" + position_message)
 
